# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import List
from attr import dataclass

from .util import SerializableAttrs, SerializableEnum
from .event import EventType
from .primitive import RoomID, UserID


class EventFormat(SerializableEnum):
    """
    Federation event format enum, as specified in the `create filter endpoint`_.

    .. _create filter endpoint:
        https://matrix.org/docs/spec/client_server/r0.5.0#post-matrix-client-r0-user-userid-filter
    """
    CLIENT = "client"
    FEDERATION = "federation"


@dataclass
class EventFilter(SerializableAttrs['EventFilter']):
    """
    Event filter object, as specified in the `create filter endpoint`_.

    Attributes:
        limit: The maximum number of events to return.
        not_senders: A list of sender IDs to exclude. If this list is absent then no senders are
            excluded. A matching sender will be excluded even if it is listed in the :attr:`senders`
            filter.
        not_types: A list of event types to exclude. If this list is absent then no event types are
            excluded. A matching type will be excluded even if it is listed in the :attr:`types`
            filter. A ``'*'`` can be used as a wildcard to match any sequence of characters.
        senders: A list of senders IDs to include. If this list is absent then all senders are
            included.
        types: A list of event types to include. If this list is absent then all event types are
            included. A ``'*'`` can be used as a wildcard to match any sequence of characters.

    .. _create filter endpoint:
        https://matrix.org/docs/spec/client_server/r0.5.0#post-matrix-client-r0-user-userid-filter
    """
    limit: int = None
    not_senders: List[UserID] = None
    not_types: List[EventType] = None
    senders: List[UserID] = None
    types: List[EventType] = None


@dataclass
class RoomEventFilter(EventFilter, SerializableAttrs['RoomEventFilter']):
    """
    Room event filter object, as specified in the `create filter endpoint`_.

    Attributes:
        lazy_load_members: If ``True``, enables lazy-loading of membership events. See `Lazy-loading
            room members`_ for more information.
        include_redundant_members: If ``True``, sends all membership events for all events,
            even if they have already been sent to the client. Does not apply unless
            :attr:`lazy_load_members` is true. See `Lazy-loading room members`_ for more
            information.
        not_rooms: A list of room IDs to exclude. If this list is absent then no rooms are excluded.
            A matching room will be excluded even if it is listed in the :attr:`rooms` filter.
        rooms: A list of room IDs to include. If this list is absent then all rooms are included.
        contains_url: If ``True``, includes only events with a url key in their content. If
            ``False``, excludes those events. If omitted, ``url`` key is not considered for
            filtering.

    .. _create filter endpoint:
        https://matrix.org/docs/spec/client_server/r0.5.0#post-matrix-client-r0-user-userid-filter
    .. _Lazy-loading room members:
        https://matrix.org/docs/spec/client_server/r0.5.0#lazy-loading-room-members
    """
    lazy_load_members: bool = False
    include_redundant_members: bool = False
    not_rooms: List[RoomID] = None
    rooms: List[RoomID] = None
    contains_url: bool = None


@dataclass
class StateFilter(RoomEventFilter, SerializableAttrs['RoomEventFilter']):
    """
    State event filter object, as specified in the `create filter endpoint`_. Currently this is the
    same as :class:`RoomEventFilter`.

    .. _create filter endpoint:
        https://matrix.org/docs/spec/client_server/r0.5.0#post-matrix-client-r0-user-userid-filter
    """
    pass


@dataclass
class RoomFilter(SerializableAttrs['RoomFilter']):
    """
    Room filter object, as specified in the `create filter endpoint`_.

    Attributes:
        not_rooms: A list of room IDs to exclude. If this list is absent then no rooms are excluded.
            A matching room will be excluded even if it is listed in the ``'rooms'`` filter. This
            filter is applied before the filters in :attr:`ephemeral`, :attr:`state`,
            :attr:`timeline` or :attr:`account_data`.
        rooms: A list of room IDs to include. If this list is absent then all rooms are included.
            This filter is applied before the filters in :attr:`ephemeral`, :attr:`state`,
            :attr:`timeline` or :attr:`account_data`.
        ephemeral: The events that aren't recorded in the room history, e.g. typing and receipts,
            to include for rooms.
        include_leave: Include rooms that the user has left in the sync.
        state: The state events to include for rooms.
        timeline: The message and state update events to include for rooms.
        account_data: The per user account data to include for rooms.

    .. _create filter endpoint:
        https://matrix.org/docs/spec/client_server/r0.5.0#post-matrix-client-r0-user-userid-filter
    """
    not_rooms: List[RoomID] = None
    rooms: List[RoomID] = None
    ephemeral: RoomEventFilter = None
    include_leave: bool = False
    state: StateFilter = None
    timeline: RoomEventFilter = None
    account_data: RoomEventFilter = None


@dataclass
class Filter(SerializableAttrs['Filter']):
    """
    Base filter object, as specified in the `create filter endpoint`_.

    Attributes:
        event_fields: List of event fields to include. If this list is absent then all fields are
            included. The entries may include ``.`` charaters to indicate sub-fields. So
            ``['content.body']`` will include the ``body`` field of the ``content`` object. A
            literal ``.`` character in a field name may be escaped using a ``\\``. A server may
            include more fields than were requested.
        event_format: The format to use for events. ``'client'`` will return the events in a format
            suitable for clients. ``'federation'`` will return the raw event as receieved over
            federation. The default is :attr:`~EventFormat.CLIENT`.
        presence: The presence updates to include.
        account_data: The user account data that isn't associated with rooms to include.
        room: Filters to be applied to room data.

    .. _create filter endpoint:
        https://matrix.org/docs/spec/client_server/r0.5.0#post-matrix-client-r0-user-userid-filter
    """
    event_fields: List[str] = None
    event_format: EventFormat = None
    presence: EventFilter = None
    account_data: EventFilter = None
    room: RoomFilter = None



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional
from enum import IntEnum

from attr import dataclass
import attr

from mautrix.types import (UserID, DeviceID, IdentityKey, SigningKey, SerializableAttrs,
                           ToDeviceEvent)


class TrustState(IntEnum):
    UNSET = 0
    VERIFIED = 1
    BLACKLISTED = 2
    IGNORED = 3


@dataclass
class DeviceIdentity:
    user_id: UserID
    device_id: DeviceID
    identity_key: IdentityKey
    signing_key: SigningKey

    trust: TrustState
    deleted: bool
    name: str


@dataclass
class OlmEventKeys(SerializableAttrs['OlmEventKeys']):
    ed25519: SigningKey


@dataclass
class DecryptedOlmEvent(ToDeviceEvent, SerializableAttrs['EncryptedOlmEvent']):
    keys: OlmEventKeys
    recipient: UserID
    recipient_keys: OlmEventKeys
    sender_device: Optional[DeviceID] = None
    sender_key: IdentityKey = attr.ib(metadata={"hidden": True}, default=None)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, List, Optional, Union, NamedTuple

from mautrix.types import (Member, MemberStateEventContent, Membership, PowerLevelStateEventContent,
                           RoomEncryptionStateEventContent, RoomID, UserID)
from mautrix.util.async_db import Database

from ..abstract import StateStore
from .upgrade import upgrade_table


class RoomState(NamedTuple):
    is_encrypted: bool
    has_full_member_list: bool
    encryption: RoomEncryptionStateEventContent
    power_levels: PowerLevelStateEventContent


class PgStateStore(StateStore):
    upgrade_table = upgrade_table

    db: Database

    def __init__(self, db: Database) -> None:
        self.db = db

    async def get_member(self, room_id: RoomID, user_id: UserID) -> Optional[Member]:
        res = await self.db.fetchrow("SELECT membership, displayname, avatar_url "
                                     "FROM mx_user_profile WHERE room_id=$1 AND user_id=$2",
                                     room_id, user_id)
        if res is None:
            return None
        return Member(membership=Membership.deserialize(res["membership"]),
                      displayname=res["displayname"], avatar_url=res["avatar_url"])

    async def set_member(self, room_id: RoomID, user_id: UserID,
                         member: Union[Member, MemberStateEventContent]) -> None:
        q = ("INSERT INTO mx_user_profile (room_id, user_id, membership, displayname, avatar_url) "
             "VALUES ($1, $2, $3, $4, $5)"
             "ON CONFLICT (room_id, user_id) DO UPDATE SET membership=$3, displayname=$4,"
             "                                             avatar_url=$5")
        await self.db.execute(q, room_id, user_id, member.membership.value,
                              member.displayname, member.avatar_url)

    async def set_membership(self, room_id: RoomID, user_id: UserID,
                             membership: Membership) -> None:
        q = ("INSERT INTO mx_user_profile (room_id, user_id, membership) VALUES ($1, $2, $3) "
             "ON CONFLICT (room_id, user_id) DO UPDATE SET membership=$3")
        await self.db.execute(q, room_id, user_id, membership.value)

    async def get_members(self, room_id: RoomID) -> Optional[List[UserID]]:
        res = await self.db.fetch("SELECT user_id FROM mx_user_profile "
                                  "WHERE room_id=$1 AND (membership='join' OR membership='invite')",
                                  room_id)
        return [profile["user_id"] for profile in res]

    async def get_members_filtered(self, room_id: RoomID, not_prefix: str, not_suffix: str,
                                   not_id: str) -> Optional[List[UserID]]:
        res = await self.db.fetch("SELECT user_id FROM mx_user_profile "
                                  "WHERE room_id=$1 AND (membership='join' OR membership='invite')"
                                  "AND user_id != $2 AND user_id NOT LIKE $3",
                                  room_id, not_id, f"{not_prefix}%{not_suffix}")
        return [profile["user_id"] for profile in res]

    async def set_members(self, room_id: RoomID,
                          members: Dict[UserID, Union[Member, MemberStateEventContent]]) -> None:
        columns = ["room_id", "user_id", "membership", "displayname", "avatar_url"]
        records = [(room_id, user_id, str(member.membership), member.displayname, member.avatar_url)
                   for user_id, member in members.items()]
        async with self.db.acquire() as conn, conn.transaction():
            del_q = "DELETE FROM mx_user_profile WHERE room_id=$1"
            await conn.execute(del_q, room_id)
            await conn.copy_records_to_table("mx_user_profile", records=records, columns=columns)

    async def has_full_member_list(self, room_id: RoomID) -> bool:
        return bool(await self.db.fetchval("SELECT has_full_member_list FROM mx_room_state "
                                           "WHERE room_id=$1", room_id))

    async def has_power_levels_cached(self, room_id: RoomID) -> bool:
        return bool(await self.db.fetchval("SELECT power_levels IS NOT NULL FROM mx_room_state "
                                           "WHERE room_id=$1", room_id))

    async def get_power_levels(self, room_id: RoomID) -> Optional[PowerLevelStateEventContent]:
        power_levels_json = await self.db.fetchval("SELECT power_levels FROM mx_room_state "
                                                   "WHERE room_id=$1", room_id)
        if power_levels_json is None:
            return None
        return PowerLevelStateEventContent.parse_json(power_levels_json)

    async def set_power_levels(self, room_id: RoomID, content: PowerLevelStateEventContent) -> None:
        await self.db.execute("INSERT INTO mx_room_state (room_id, power_levels) VALUES ($1, $2) "
                              "ON CONFLICT (room_id) DO UPDATE SET power_levels=$2",
                              room_id, content.json())

    async def has_encryption_info_cached(self, room_id: RoomID) -> bool:
        return bool(await self.db.fetchval("SELECT encryption IS NULL FROM mx_room_state "
                                           "WHERE room_id=$1", room_id))

    async def is_encrypted(self, room_id: RoomID) -> Optional[bool]:
        return await self.db.fetchval("SELECT is_encrypted FROM mx_room_state WHERE room_id=$1",
                                      room_id)

    async def get_encryption_info(self, room_id: RoomID
                                  ) -> Optional[RoomEncryptionStateEventContent]:
        row = await self.db.fetchrow("SELECT is_encrypted, encryption FROM mx_room_state "
                                     "WHERE room_id=$1", room_id)
        if row is None or not row["is_encrypted"]:
            return None
        return RoomEncryptionStateEventContent.parse_json(row["encryption"])

    async def set_encryption_info(self, room_id: RoomID,
                                  content: RoomEncryptionStateEventContent) -> None:
        q = ("INSERT INTO mx_room_state (room_id, is_encrypted, encryption) VALUES ($1, true, $2) "
             "ON CONFLICT (room_id) DO UPDATE SET is_encrypted=true, encryption=$2")
        await self.db.execute(q, room_id, content.json())



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Tuple, Optional, Union, TYPE_CHECKING
import logging
import asyncio
import os.path
import time
import sys

from yarl import URL

from mautrix.types import (EventID, RoomID, UserID, Event, EventType, MessageEvent, MessageType,
                           MessageEventContent, StateEvent, Membership, MemberStateEventContent,
                           PresenceEvent, TypingEvent, ReceiptEvent, TextMessageEventContent,
                           EncryptedEvent, ReceiptType, SingleReceiptEventContent, StateUnsigned,
                           MediaRepoConfig)
from mautrix.errors import IntentError, MatrixError, MForbidden, DecryptionError, SessionNotFound
from mautrix.appservice import AppService
from mautrix.util.logging import TraceLogger
from mautrix.util.opt_prometheus import Histogram

from .commands import CommandProcessor

if TYPE_CHECKING:
    from .config import BaseBridgeConfig
    from .user import BaseUser
    from .portal import BasePortal
    from .puppet import BasePuppet
    from .bridge import Bridge

try:
    from .e2ee import EncryptionManager
except ImportError:
    EncryptionManager = None

EVENT_TIME = Histogram("bridge_matrix_event", "Time spent processing Matrix events", ["event_type"])


class BaseMatrixHandler:
    log: TraceLogger = logging.getLogger("mau.mx")
    az: AppService
    commands: CommandProcessor
    config: 'BaseBridgeConfig'
    bridge: 'Bridge'
    e2ee: Optional[EncryptionManager]
    media_config: MediaRepoConfig

    user_id_prefix: str
    user_id_suffix: str

    def __init__(self, command_processor: Optional[CommandProcessor] = None,
                 bridge: Optional['Bridge'] = None) -> None:
        self.az = bridge.az
        self.config = bridge.config
        self.bridge = bridge
        self.commands = command_processor or CommandProcessor(bridge=bridge)
        self.media_config = MediaRepoConfig(upload_size=50 * 1024 * 1024)
        self.az.matrix_event_handler(self.int_handle_event)

        self.e2ee = None
        if self.config["bridge.encryption.allow"]:
            if not EncryptionManager:
                self.log.error("Encryption enabled in config, but dependencies not installed.")
                return
            self.e2ee = EncryptionManager(
                bridge=bridge,
                user_id_prefix=self.user_id_prefix, user_id_suffix=self.user_id_suffix,
                homeserver_address=self.config["homeserver.address"],
                db_url=self._get_db_url(bridge.config),
                key_sharing_config=self.config["bridge.encryption.key_sharing"])

    @staticmethod
    def _get_db_url(config: 'BaseBridgeConfig') -> str:
        db_url = config["bridge.encryption.database"]
        if not db_url or db_url == "default":
            db_url = config["appservice.database"]
            parsed_url = URL(db_url)
            if parsed_url.scheme == "sqlite":
                # Remove the extension to replace it with our own
                path = os.path.splitext(parsed_url.path)[0]
                db_url = f"pickle://{path}.crypto.pickle"
        return db_url

    async def wait_for_connection(self) -> None:
        self.log.info("Ensuring connectivity to homeserver")
        errors = 0
        tried_to_register = False
        while True:
            try:
                await self.az.intent.whoami()
                break
            except MForbidden as e:
                if not tried_to_register:
                    self.log.debug("Whoami endpoint returned M_FORBIDDEN, "
                                   "trying to register bridge bot before retrying...")
                    await self.az.intent.ensure_registered()
                    tried_to_register = True
                else:
                    raise
            except Exception:
                errors += 1
                if errors <= 6:
                    self.log.exception("Connection to homeserver failed, retrying in 10 seconds")
                    await asyncio.sleep(10)
                else:
                    raise
        try:
            self.media_config = await self.az.intent.get_media_repo_config()
        except Exception:
            self.log.warning("Failed to fetch media repo config", exc_info=True)

    async def init_as_bot(self) -> None:
        self.log.debug("Initializing appservice bot")
        displayname = self.config["appservice.bot_displayname"]
        if displayname:
            try:
                await self.az.intent.set_displayname(
                    displayname if displayname != "remove" else "")
            except Exception:
                self.log.exception("Failed to set bot displayname")

        avatar = self.config["appservice.bot_avatar"]
        if avatar:
            try:
                await self.az.intent.set_avatar_url(avatar if avatar != "remove" else "")
            except Exception:
                self.log.exception("Failed to set bot avatar")

    async def init_encryption(self) -> None:
        if self.e2ee:
            if not await self.e2ee.check_server_support():
                self.log.critical("Encryption enabled in config, but homeserver does not advertise "
                                  "appservice login")
                sys.exit(30)
            await self.e2ee.start()

    @staticmethod
    async def allow_message(user: 'BaseUser') -> bool:
        return user.is_whitelisted

    @staticmethod
    async def allow_command(user: 'BaseUser') -> bool:
        return user.is_whitelisted

    @staticmethod
    async def allow_bridging_message(user: 'BaseUser', portal: 'BasePortal') -> bool:
        return await user.is_logged_in()

    async def handle_leave(self, room_id: RoomID, user_id: UserID, event_id: EventID) -> None:
        pass

    async def handle_kick(self, room_id: RoomID, user_id: UserID, kicked_by: UserID, reason: str,
                          event_id: EventID) -> None:
        pass

    async def handle_ban(self, room_id: RoomID, user_id: UserID, banned_by: UserID, reason: str,
                         event_id: EventID) -> None:
        pass

    async def handle_unban(self, room_id: RoomID, user_id: UserID, unbanned_by: UserID,
                           reason: str, event_id: EventID) -> None:
        pass

    async def handle_join(self, room_id: RoomID, user_id: UserID, event_id: EventID) -> None:
        pass

    async def handle_member_info_change(self, room_id: RoomID, user_id: UserID,
                                        content: MemberStateEventContent,
                                        prev_content: MemberStateEventContent,
                                        event_id: EventID) -> None:
        pass

    async def handle_puppet_invite(self, room_id: RoomID, puppet: 'BasePuppet',
                                   invited_by: 'BaseUser', event_id: EventID) -> None:
        pass

    async def handle_invite(self, room_id: RoomID, user_id: UserID, inviter: 'BaseUser',
                            event_id: EventID) -> None:
        pass

    async def handle_reject(self, room_id: RoomID, user_id: UserID, reason: str, event_id: EventID
                            ) -> None:
        pass

    async def handle_disinvite(self, room_id: RoomID, user_id: UserID, disinvited_by: UserID,
                               reason: str, event_id: EventID) -> None:
        pass

    async def handle_event(self, evt: Event) -> None:
        """Called by :meth:`int_handle_event` for message events other than m.room.message."""

    async def handle_state_event(self, evt: StateEvent) -> None:
        """Called by :meth:`int_handle_event` for state events other than m.room.membership."""

    async def handle_ephemeral_event(self, evt: Union[ReceiptEvent, PresenceEvent, TypingEvent]
                                     ) -> None:
        if evt.type == EventType.RECEIPT:
            await self.handle_receipt(evt)

    async def send_permission_error(self, room_id: RoomID) -> None:
        await self.az.intent.send_notice(
            room_id,
            text="You are not whitelisted to use this bridge.\n\n"
                 "If you are the owner of this bridge, see the bridge.permissions "
                 "section in your config file.",
            html="<p>You are not whitelisted to use this bridge.</p>"
                 "<p>If you are the owner of this bridge, see the "
                 "<code>bridge.permissions</code> section in your config file.</p>")

    async def accept_bot_invite(self, room_id: RoomID, inviter: 'BaseUser') -> None:
        tries = 0
        while tries < 5:
            try:
                await self.az.intent.join_room(room_id)
                break
            except (IntentError, MatrixError):
                tries += 1
                wait_for_seconds = (tries + 1) * 10
                if tries < 5:
                    self.log.exception(f"Failed to join room {room_id} with bridge bot, "
                                       f"retrying in {wait_for_seconds} seconds...")
                    await asyncio.sleep(wait_for_seconds)
                else:
                    self.log.exception(f"Failed to join room {room_id}, giving up.")
                    return

        if not await self.allow_command(inviter):
            await self.send_permission_error(room_id)
            await self.az.intent.leave_room(room_id)
            return

        await self.send_welcome_message(room_id, inviter)

    async def send_welcome_message(self, room_id: RoomID, inviter: 'BaseUser') -> None:
        await self.az.intent.send_notice(
            room_id=room_id,
            text=f"Hello, I'm a bridge bot. Use `{self.commands.command_prefix} help` for help.",
            html=f"Hello, I'm a bridge bot. Use <code>{self.commands.command_prefix} help</code> "
                 f"for help.")

    async def int_handle_invite(self, room_id: RoomID, user_id: UserID, invited_by: UserID,
                                event_id: EventID) -> None:
        self.log.debug(f"{invited_by} invited {user_id} to {room_id}")
        inviter = await self.bridge.get_user(invited_by)
        if inviter is None:
            self.log.exception(f"Failed to find user with Matrix ID {invited_by}")
            return
        elif user_id == self.az.bot_mxid:
            await self.accept_bot_invite(room_id, inviter)
            return
        elif not await self.allow_command(inviter):
            return

        puppet = await self.bridge.get_puppet(user_id)
        if puppet:
            await self.handle_puppet_invite(room_id, puppet, inviter, event_id)
            return

        await self.handle_invite(room_id, user_id, inviter, event_id)

    def is_command(self, message: MessageEventContent) -> Tuple[bool, str]:
        text = message.body
        prefix = self.config["bridge.command_prefix"]
        is_command = text.startswith(prefix)
        if is_command:
            text = text[len(prefix) + 1:].lstrip()
        return is_command, text

    async def handle_message(self, room_id: RoomID, user_id: UserID, message: MessageEventContent,
                             event_id: EventID) -> None:
        sender = await self.bridge.get_user(user_id)
        if not sender or not await self.allow_message(sender):
            self.log.debug(f"Ignoring message {event_id} from {user_id} to {room_id}:"
                           " User is not whitelisted.")
            return
        self.log.debug(f"Received Matrix event {event_id} from {sender.mxid} in {room_id}")
        self.log.trace("Event %s content: %s", event_id, message)

        if isinstance(message, TextMessageEventContent):
            message.trim_reply_fallback()

        is_command, text = self.is_command(message)
        portal = await self.bridge.get_portal(room_id)
        if not is_command and portal and await self.allow_bridging_message(sender, portal):
            await portal.handle_matrix_message(sender, message, event_id)
            return

        if message.msgtype != MessageType.TEXT:
            return
        elif not await self.allow_command(sender):
            self.log.trace("Ignoring command %s from %s", event_id, sender.mxid)
            return

        has_two_members, bridge_bot_in_room = await self._is_direct_chat(room_id)
        is_management = has_two_members and bridge_bot_in_room

        if is_command or is_management:
            try:
                command, arguments = text.split(" ", 1)
                args = arguments.split(" ")
            except ValueError:
                # Not enough values to unpack, i.e. no arguments
                command = text
                args = []
            await self.commands.handle(room_id, event_id, sender, command, args, message, portal,
                                       is_management, bridge_bot_in_room)
        else:
            self.log.trace("Ignoring event %s from %s: not a command and not a portal room",
                           event_id, sender.mxid)

    async def _is_direct_chat(self, room_id: RoomID) -> Tuple[bool, bool]:
        try:
            members = await self.az.intent.get_room_members(room_id)
            return len(members) == 2, self.az.bot_mxid in members
        except MatrixError:
            return False, False

    async def handle_receipt(self, evt: ReceiptEvent) -> None:
        for event_id, receipts in evt.content.items():
            for user_id, data in receipts[ReceiptType.READ].items():
                user = await self.bridge.get_user(user_id, create=False)
                if not user or not await user.is_logged_in():
                    continue

                portal = await self.bridge.get_portal(evt.room_id)
                if not portal:
                    continue

                await self.handle_read_receipt(user, portal, event_id, data)

    async def handle_read_receipt(self, user: 'BaseUser', portal: 'BasePortal', event_id: EventID,
                                  data: SingleReceiptEventContent) -> None:
        pass

    def filter_matrix_event(self, evt: Event) -> bool:
        if not isinstance(evt, (MessageEvent, StateEvent, ReceiptEvent)):
            return False
        return evt.sender == self.az.bot_mxid

    async def try_handle_sync_event(self, evt: Event) -> None:
        try:
            if isinstance(evt, (ReceiptEvent, PresenceEvent, TypingEvent)):
                await self.handle_ephemeral_event(evt)
            else:
                self.log.trace("Unknown event type received from sync: %s", evt)
        except Exception:
            self.log.exception("Error handling manually received Matrix event")

    async def send_encryption_error_notice(self, evt: EncryptedEvent,
                                           error: DecryptionError) -> None:
        await self.az.intent.send_notice(evt.room_id,
                                         f"\u26a0 Your message was not bridged: {error}")

    async def handle_encrypted(self, evt: EncryptedEvent) -> None:
        if not self.e2ee:
            await self.handle_encrypted_unsupported(evt)
            return
        try:
            decrypted = await self.e2ee.decrypt(evt, wait_session_timeout=5)
        except SessionNotFound as e:
            await self._handle_encrypted_wait(evt, e, wait=10)
        except DecryptionError as e:
            self.log.warning("Failed to decrypt event %s: %s", evt.event_id, e)
            self.log.trace("%s decryption traceback:", evt.event_id, exc_info=True)
            await self.send_encryption_error_notice(evt, e)
        else:
            await self.int_handle_event(decrypted)

    async def handle_encrypted_unsupported(self, evt: EncryptedEvent) -> None:
        self.log.debug("Got encrypted message %s from %s, but encryption is not enabled",
                       evt.event_id, evt.sender)
        await self.az.intent.send_notice(evt.room_id, "🔒️ This bridge has not been configured "
                                                      "to support encryption")

    async def _handle_encrypted_wait(self, evt: EncryptedEvent, err: SessionNotFound, wait: int
                                     ) -> None:
        self.log.warning(f"Didn't find session {err.session_id}, waiting even longer")
        msg = ("\u26a0 Your message was not bridged: the bridge hasn't received the decryption "
               f"keys. The bridge will retry for {wait} seconds. If this error keeps happening, "
               "try restarting your client.")
        event_id = await self.az.intent.send_notice(evt.room_id, msg)
        got_keys = await self.e2ee.crypto.wait_for_session(evt.room_id, err.sender_key,
                                                           err.session_id, timeout=wait)
        if got_keys:
            try:
                decrypted = await self.e2ee.decrypt(evt, wait_session_timeout=0)
            except DecryptionError as e:
                self.log.warning(f"Didn't get {err.session_id}, giving up on {evt.event_id}")
                msg = f"\u26a0 Your message was not bridged: {e}"
            else:
                await self.az.intent.redact(evt.room_id, event_id)
                await self.int_handle_event(decrypted)
                return
        else:
            msg = ("\u26a0 Your message was not bridged: the bridge hasn't received the decryption "
                   "keys. If this error keeps happening, try restarting your client.")
        content = TextMessageEventContent(msgtype=MessageType.NOTICE, body=msg)
        content.set_edit(event_id)
        await self.az.intent.send_message(evt.room_id, content)

    async def handle_encryption(self, evt: StateEvent) -> None:
        await self.az.state_store.set_encryption_info(evt.room_id, evt.content)
        portal = await self.bridge.get_portal(evt.room_id)
        if portal:
            portal.encrypted = True
            await portal.save()
            # TODO replace with normal attribute access in v0.9
            if getattr(portal, "is_direct", False):
                portal.log.debug("Received encryption event in direct portal: %s", evt.content)
                await portal.enable_dm_encryption()

    async def int_handle_event(self, evt: Event) -> None:
        if isinstance(evt, StateEvent) and evt.type == EventType.ROOM_MEMBER and self.e2ee:
            await self.e2ee.handle_member_event(evt)
        if self.filter_matrix_event(evt):
            return
        self.log.trace("Received event: %s", evt)
        start_time = time.time()

        if evt.type == EventType.ROOM_MEMBER:
            evt: StateEvent
            unsigned = evt.unsigned or StateUnsigned()
            prev_content = unsigned.prev_content or MemberStateEventContent()
            prev_membership = prev_content.membership if prev_content else Membership.JOIN
            if evt.content.membership == Membership.INVITE:
                await self.int_handle_invite(evt.room_id, UserID(evt.state_key), evt.sender,
                                             evt.event_id)
            elif evt.content.membership == Membership.LEAVE:
                if prev_membership == Membership.BAN:
                    await self.handle_unban(evt.room_id, UserID(evt.state_key), evt.sender,
                                            evt.content.reason, evt.event_id)
                elif prev_membership == Membership.INVITE:
                    if evt.sender == evt.state_key:
                        await self.handle_reject(evt.room_id, UserID(evt.state_key),
                                                 evt.content.reason, evt.event_id)
                    else:
                        await self.handle_disinvite(evt.room_id, UserID(evt.state_key), evt.sender,
                                                    evt.content.reason, evt.event_id)
                elif evt.sender == evt.state_key:
                    await self.handle_leave(evt.room_id, UserID(evt.state_key), evt.event_id)
                else:
                    await self.handle_kick(evt.room_id, UserID(evt.state_key), evt.sender,
                                           evt.content.reason, evt.event_id)
            elif evt.content.membership == Membership.BAN:
                await self.handle_ban(evt.room_id, UserID(evt.state_key), evt.sender,
                                      evt.content.reason, evt.event_id)
            elif evt.content.membership == Membership.JOIN:
                if prev_membership != Membership.JOIN:
                    await self.handle_join(evt.room_id, UserID(evt.state_key), evt.event_id)
                else:
                    await self.handle_member_info_change(evt.room_id, UserID(evt.state_key),
                                                         evt.content, prev_content, evt.event_id)
        elif evt.type in (EventType.ROOM_MESSAGE, EventType.STICKER):
            evt: MessageEvent
            if evt.type != EventType.ROOM_MESSAGE:
                evt.content.msgtype = MessageType(str(evt.type))
            await self.handle_message(evt.room_id, evt.sender, evt.content, evt.event_id)
        elif evt.type == EventType.ROOM_ENCRYPTED:
            await self.handle_encrypted(evt)
        elif evt.type == EventType.ROOM_ENCRYPTION:
            await self.handle_encryption(evt)
        else:
            if evt.type.is_state and isinstance(evt, StateEvent):
                await self.handle_state_event(evt)
            elif evt.type.is_ephemeral and isinstance(evt, (PresenceEvent, TypingEvent,
                                                            ReceiptEvent)):
                await self.handle_ephemeral_event(evt)
            else:
                await self.handle_event(evt)

        await self.log_event_handle_duration(evt, time.time() - start_time)

    async def log_event_handle_duration(self, evt: Event, duration: float) -> None:
        EVENT_TIME.labels(event_type=str(evt.type)).observe(duration)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import TYPE_CHECKING, IO, Any, Union, Optional
from abc import ABC, abstractmethod
from pathlib import Path
import pickle
import json
import time

if TYPE_CHECKING:
    from typing import Protocol


    class Filer(Protocol):
        def dump(self, obj: Any, file: IO) -> None: ...

        def load(self, file: IO) -> Any: ...

PathLike = Union[str, Path, IO]


class FileStore(ABC):
    path: PathLike
    filer: 'Filer'
    binary: bool
    save_interval: float
    _last_save: float

    def __init__(self, path: PathLike, filer: Optional['Filer'] = None, binary: bool = True,
                 save_interval: float = 60.0) -> None:
        self.path = path
        self.filer = filer or (pickle if binary else json)
        self.binary = binary
        self.save_interval = save_interval
        self._last_save = time.monotonic()

    @abstractmethod
    def serialize(self) -> Any:
        pass

    @abstractmethod
    def deserialize(self, data: Any) -> None:
        pass

    def _save(self) -> None:
        if isinstance(self.path, IO):
            file = self.path
            close = False
        else:
            file = open(self.path, "wb" if self.binary else "w")
            close = True
        try:
            self.filer.dump(self.serialize(), file)
        finally:
            if close:
                file.close()

    def _load(self) -> None:
        if isinstance(self.path, IO):
            file = self.path
            close = False
        else:
            try:
                file = open(self.path, "rb" if self.binary else "r")
            except FileNotFoundError:
                return
            close = True
        try:
            self.deserialize(self.filer.load(file))
        finally:
            if close:
                file.close()

    async def flush(self) -> None:
        self._save()

    async def open(self) -> None:
        self._load()

    def _time_limited_flush(self) -> None:
        if self._last_save + self.save_interval < time.monotonic():
            self._save()
            self._last_save = time.monotonic()



# Copyright (c) 2021 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Any, Callable, Awaitable
import functools


def async_getter_lock(fn: Callable[[Any, Any], Awaitable[Any]]) -> Any:
    @functools.wraps(fn)
    async def wrapper(cls, *args, **kwargs) -> Any:
        async with cls._async_get_locks[args]:
            return await fn(cls, *args, **kwargs)

    return wrapper



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import List, NewType, NamedTuple, Dict
from enum import Enum

from attr import dataclass
import attr

from .primitive import RoomID, RoomAlias, SyncToken, ContentURI, UserID
from .util import SerializableAttrs
from .event import Event

DeviceLists = NamedTuple("DeviceLists", changed=List[UserID], left=List[UserID])
DeviceOTKCount = NamedTuple("DeviceOTKCount", curve25519=int, signed_curve25519=int)


class RoomCreatePreset(Enum):
    """
    Room creation preset, as specified in the `createRoom endpoint`_

    .. _createRoom endpoint:
        https://matrix.org/docs/spec/client_server/r0.5.0#post-matrix-client-r0-createroom
    """
    PRIVATE = "private_chat"
    TRUSTED_PRIVATE = "trusted_private_chat"
    PUBLIC = "public_chat"


class RoomDirectoryVisibility(Enum):
    """
    Room directory visibility, as specified in the `createRoom endpoint`_

    .. _createRoom endpoint:
        https://matrix.org/docs/spec/client_server/r0.5.0#post-matrix-client-r0-createroom
    """
    PRIVATE = "private"
    PUBLIC = "public"


class PaginationDirection(Enum):
    """
    Pagination direction, as specified in the `pagination section`_.

    .. _pagination section:
        https://matrix.org/docs/spec/client_server/latest#pagination
    """
    FORWARD = "f"
    BACKWARD = "b"


@dataclass
class RoomAliasInfo(SerializableAttrs['RoomAliasInfo']):
    """
    Room alias query result, as specified in the `alias resolve endpoint`_

    Attributes:
        room_id: The room ID for this room alias.
        servers: A list of servers that are aware of this room alias.

    .. _alias resolve endpoint:
        https://matrix.org/docs/spec/client_server/r0.5.0#get-matrix-client-r0-directory-room-roomalias
    """
    room_id: RoomID = None
    servers: List[str] = None


DirectoryPaginationToken = NewType("DirectoryPaginationToken", str)


@dataclass
class PublicRoomInfo(SerializableAttrs['PublicRoomInfo']):
    room_id: RoomID

    num_joined_members: int

    world_readable: bool
    guests_can_join: bool

    name: str = None
    topic: str = None
    avatar_url: ContentURI = None

    aliases: List[RoomAlias] = None
    canonical_alias: RoomAlias = None


@dataclass
class RoomDirectoryResponse(SerializableAttrs['RoomDirectoryResponse']):
    chunk: List[PublicRoomInfo]
    next_batch: DirectoryPaginationToken = None
    prev_batch: DirectoryPaginationToken = None
    total_room_count_estimate: int = None


PaginatedMessages = NamedTuple("PaginatedMessages", start=SyncToken, end=SyncToken,
                               events=List[Event])


@dataclass
class VersionsResponse(SerializableAttrs['VersionsResponse']):
    versions: List[str]
    unstable_features: Dict[str, bool] = attr.ib(factory=lambda: {})



import setuptools

from mautrix import __version__

setuptools.setup(
    name="mautrix",
    version=__version__,
    url="https://github.com/tulir/mautrix-python",

    author="Tulir Asokan",
    author_email="tulir@maunium.net",

    description="A Python 3 asyncio Matrix framework.",
    long_description=open("README.rst").read(),

    packages=setuptools.find_packages(),

    install_requires=[
        "aiohttp>=3,<4",
        "attrs>=18.1.0",
        "yarl>=1,<2",
    ],
    extras_require={
        "detect_mimetype": ["python-magic>=0.4.15,<0.5"],
    },
    python_requires="~=3.6",

    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Topic :: Communications :: Chat",
        "Framework :: AsyncIO",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],

    package_data={
        "mautrix.types.event": ["type.pyi"],
        "mautrix.util": ["opt_prometheus.pyi"],
        "mautrix.util.formatter": ["html_reader.pyi"],
    },
)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import TYPE_CHECKING

from .base import MatrixError

if TYPE_CHECKING:
    from mautrix.types import SessionID, IdentityKey


class CryptoError(MatrixError):
    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message


class EncryptionError(CryptoError):
    pass


class SessionShareError(CryptoError):
    pass


class DecryptionError(CryptoError):
    pass


class MatchingSessionDecryptionError(DecryptionError):
    pass


class SessionNotFound(DecryptionError):
    def __init__(self, session_id: 'SessionID', sender_key: 'IdentityKey') -> None:
        super().__init__("Failed to decrypt megolm event: "
                         f"no session with given ID {session_id} found")
        self.session_id = session_id
        self.sender_key = sender_key


class DuplicateMessageIndex(DecryptionError):
    def __init__(self) -> None:
        super().__init__("Duplicate message index")


class VerificationError(DecryptionError):
    def __init__(self) -> None:
        super().__init__("Device keys in event and verified device info do not match")


class DecryptedPayloadError(DecryptionError):
    pass


class MismatchingRoomError(DecryptionError):
    def __init__(self) -> None:
        super().__init__("Encrypted megolm event is not intended for this room")


class DeviceValidationError(EncryptionError):
    pass



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Union, Optional, List, NewType
from attr import dataclass
import attr

from .primitive import UserID, DeviceID, JSON
from .util import SerializableAttrs, ExtensibleEnum, deserializer, Obj


class LoginType(ExtensibleEnum):
    """
    A login type, as specified in the `POST /login endpoint`_

    .. _POST /login endpoint:
        https://matrix.org/docs/spec/client_server/r0.6.1#post-matrix-client-r0-login
    """
    PASSWORD: 'LoginType' = "m.login.password"
    TOKEN: 'LoginType' = "m.login.token"
    SSO: 'LoginType' = "m.login.sso"

    JWT: 'LoginType' = "org.matrix.login.jwt"
    APPSERVICE: 'LoginType' = "uk.half-shot.msc2778.login.application_service"


@dataclass
class LoginFlow(SerializableAttrs['LoginFlow']):
    """
    A login flow, as specified in the `GET /login endpoint`_

    .. _GET /login endpoint:
        https://matrix.org/docs/spec/client_server/r0.6.1#get-matrix-client-r0-login
    """
    type: LoginType


@dataclass
class LoginFlowList(SerializableAttrs['LoginFlowList']):
    flows: List[LoginFlow]

    def supports_type(self, type: LoginType) -> bool:
        for flow in self.flows:
            if flow.type == type:
                return True
        return False


class UserIdentifierType(ExtensibleEnum):
    """
    A user identifier type, as specified in the `Identifier types`_ section of the login spec.

    .. _Identifier types:
        https://matrix.org/docs/spec/client_server/latest#identifier-types
    """
    MATRIX_USER: 'UserIdentifierType' = "m.id.user"
    THIRD_PARTY: 'UserIdentifierType' = "m.id.thirdparty"
    PHONE: 'UserIdentifierType' = "m.id.phone"


@dataclass
class MatrixUserIdentifier(SerializableAttrs['MatrixUserIdentifier']):
    """
    A client can identify a user using their Matrix ID. This can either be the fully qualified
    Matrix user ID, or just the localpart of the user ID.

    Attributes:
        user: The Matrix user ID or localpart
    """

    user: str
    type: UserIdentifierType = UserIdentifierType.MATRIX_USER


@dataclass
class ThirdPartyIdentifier(SerializableAttrs['ThirdPartyIdentifier']):
    """
    A client can identify a user using a 3PID associated with the user's account on the homeserver,
    where the 3PID was previously associated using the `/account/3pid`_ API. See the `3PID Types`_
    Appendix for a list of Third-party ID media.

    .. _/account/3pid:
        https://matrix.org/docs/spec/client_server/latest#post-matrix-client-r0-account-3pid
    .. _3PID Types:
        https://matrix.org/docs/spec/appendices.html#pid-types
    """
    medium: str
    address: str
    type: UserIdentifierType = UserIdentifierType.THIRD_PARTY


@dataclass
class PhoneIdentifier(SerializableAttrs['PhoneIdentifier']):
    """
    A client can identify a user using a phone number associated with the user's account, where the
    phone number was previously associated using the `/account/3pid`_ API. The phone number can be
    passed in as entered by the user; the homeserver will be responsible for canonicalising it.
    If the client wishes to canonicalise the phone number, then it can use the ``m.id.thirdparty``
    identifier type with a ``medium`` of ``msisdn`` instead.

    .. _/account/3pid:
        https://matrix.org/docs/spec/client_server/latest#post-matrix-client-r0-account-3pid
    """
    country: str
    phone: str
    type: UserIdentifierType = UserIdentifierType.PHONE


UserIdentifier = NewType('UserIdentifier',
                         Union[MatrixUserIdentifier, ThirdPartyIdentifier, PhoneIdentifier])


@deserializer(UserIdentifier)
def deserialize_user_identifier(data: JSON) -> Union[UserIdentifier, Obj]:
    try:
        identifier_type = UserIdentifierType.deserialize(data["type"])
    except KeyError:
        return Obj(**data)
    if identifier_type == UserIdentifierType.MATRIX_USER:
        return MatrixUserIdentifier.deserialize(data)
    elif identifier_type == UserIdentifierType.THIRD_PARTY:
        return ThirdPartyIdentifier.deserialize(data)
    elif identifier_type == UserIdentifierType.PHONE:
        return PhoneIdentifier.deserialize(data)
    else:
        return Obj(**data)


setattr(UserIdentifier, "deserialize", deserialize_user_identifier)


@dataclass
class DiscoveryServer(SerializableAttrs['DiscoveryServer']):
    base_url: Optional[str] = None


@dataclass
class DiscoveryIntegrationServer(SerializableAttrs['DiscoveryIntegrations']):
    ui_url: Optional[str] = None
    api_url: Optional[str] = None


@dataclass
class DiscoveryIntegrations(SerializableAttrs['DiscoveryIntegrations']):
    managers: List[DiscoveryIntegrationServer] = attr.ib(factory=lambda: [])


@dataclass
class DiscoveryInformation(SerializableAttrs['DiscoveryInformation']):
    homeserver: Optional[DiscoveryServer] = attr.ib(metadata={"json": "m.homeserver"},
                                                    factory=DiscoveryServer)
    identity_server: Optional[DiscoveryServer] = attr.ib(metadata={"json": "m.identity_server"},
                                                         factory=DiscoveryServer)
    integrations: Optional[DiscoveryServer] = attr.ib(metadata={"json": "m.integrations"},
                                                      factory=DiscoveryIntegrations)


@dataclass
class LoginResponse(SerializableAttrs['LoginResponse']):
    """
    The response for a login request, as specified in the `POST /login endpoint`_

    .. POST /login endpoint:
        https://matrix.org/docs/spec/client_server/r0.6.1#post-matrix-client-r0-login
    """
    user_id: UserID
    device_id: DeviceID
    access_token: str
    well_known: DiscoveryInformation = attr.ib(factory=DiscoveryInformation)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, Type
import logging

from mautrix.types import RoomID, UserID
from mautrix.appservice import IntentAPI
from mautrix.api import Method, Path, PathBuilder
from mautrix.util.logging import TraceLogger

from .puppet import BasePuppet
from .user import BaseUser


class NotificationDisabler:
    puppet_cls: Type[BasePuppet]
    config_enabled: bool = False
    log: TraceLogger = logging.getLogger("mau.notification_disabler")

    user_id: UserID
    room_id: RoomID
    intent: Optional[IntentAPI]
    enabled: bool

    def __init__(self, room_id: RoomID, user: BaseUser) -> None:
        self.user_id = user.mxid
        self.room_id = room_id
        self.enabled = False

    @property
    def _path(self) -> PathBuilder:
        return Path.pushrules["global"].override[("net.maunium.silence_while_backfilling"
                                                  f":{self.room_id}")]

    @property
    def _rule(self) -> dict:
        return {
            "actions": ["dont_notify"],
            "conditions": [{
                "kind": "event_match",
                "key": "room_id",
                "pattern": self.room_id,
            }]
        }

    async def __aenter__(self) -> None:
        puppet = await self.puppet_cls.get_by_custom_mxid(self.user_id)
        self.intent = puppet.intent if puppet and puppet.is_real_user else None
        if not self.intent or not self.config_enabled:
            return
        self.enabled = True
        try:
            self.log.debug(f"Disabling notifications in {self.room_id} for {self.intent.mxid}")
            await self.intent.api.request(Method.PUT, self._path, content=self._rule)
        except Exception:
            self.log.warning(f"Failed to disable notifications in {self.room_id} "
                             f"for {self.intent.mxid} while backfilling", exc_info=True)
            raise

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        if not self.enabled:
            return
        try:
            self.log.debug(f"Re-enabling notifications in {self.room_id} for {self.intent.mxid}")
            await self.intent.api.request(Method.DELETE, self._path)
        except Exception:
            self.log.warning(f"Failed to re-enable notifications in {self.room_id} "
                             f"for {self.intent.mxid} after backfilling", exc_info=True)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import List, Dict, Any, Optional, Union, Iterable

from mautrix.errors import MatrixResponseError
from mautrix.api import Method, Path
from mautrix.types import (UserID, DeviceID, EncryptionKeyAlgorithm, ClaimKeysResponse, SyncToken,
                           QueryKeysResponse, EventType, ToDeviceEventContent, Serializable)

from ..base import BaseClientAPI


class CryptoMethods(BaseClientAPI):
    """
    Methods in section `13.9 Send-to-Device messaging <https://matrix.org/docs/spec/client_server/r0.6.1#id70>`__
    and `13.11 End-to-End Encryption of the spec <https://matrix.org/docs/spec/client_server/r0.6.1#id76>`__.
    """

    async def send_to_device(self, event_type: EventType,
                             messages: Dict[UserID, Dict[DeviceID, ToDeviceEventContent]]) -> None:
        """
        Send to-device events to a set of client devices.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.6.1#put-matrix-client-r0-sendtodevice-eventtype-txnid>`__

        Args:
            event_type: The type of event to send.
            messages: The messages to send. A map from user ID, to a map from device ID to message
                body. The device ID may also be *, meaning all known devices for the user.
        """
        if not event_type.is_to_device:
            raise ValueError("Event type must be a to-device event type")
        await self.api.request(Method.PUT, Path.sendToDevice[event_type][self.api.get_txn_id()], {
            "messages": {
                user_id: {device_id: (content.serialize() if isinstance(content, Serializable)
                                      else content)
                          for device_id, content in devices.items()}
                for user_id, devices in messages.items()
            },
        })

    async def send_to_one_device(self, event_type: EventType, user_id: UserID, device_id: DeviceID,
                                 message: ToDeviceEventContent) -> None:
        """
        Send a to-device event to a single device.

        Args:
            event_type: The type of event to send.
            user_id: The user whose device to send the event to.
            device_id: The device ID to send the event to.
            message: The event content to send.
        """
        return await self.send_to_device(event_type, {user_id: {device_id: message}})

    async def upload_keys(self, one_time_keys: Optional[Dict[str, Any]] = None,
                          device_keys: Optional[Dict[str, Any]] = None,
                          ) -> Dict[EncryptionKeyAlgorithm, int]:
        """
        Publishes end-to-end encryption keys for the device.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.6.1#post-matrix-client-r0-keys-upload>`__

        Args:
            one_time_keys: One-time public keys for "pre-key" messages. The names of the properties
                should be in the format ``<algorithm>:<key_id>``. The format of the key is
                determined by the key algorithm.
            device_keys: Identity keys for the device. May be absent if no new identity keys are
                required.

        Returns:
            For each key algorithm, the number of unclaimed one-time keys of that type currently
            held on the server for this device.
        """
        data = {}
        if device_keys:
            data["device_keys"] = device_keys
        if one_time_keys:
            data["one_time_keys"] = one_time_keys
        resp = await self.api.request(Method.POST, Path.keys.upload, data)
        try:
            return {EncryptionKeyAlgorithm.deserialize(alg): count
                    for alg, count in resp["one_time_key_counts"].items()}
        except KeyError as e:
            raise MatrixResponseError("`one_time_key_counts` not in response.") from e
        except AttributeError as e:
            raise MatrixResponseError("Invalid `one_time_key_counts` field in response.") from e

    async def query_keys(self, device_keys: Union[Iterable[UserID], Dict[UserID, List[DeviceID]]],
                         token: SyncToken = "", timeout: int = 10000) -> QueryKeysResponse:
        """
        Fetch devices and their identity keys for the given users.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.6.1#post-matrix-client-r0-keys-query>`__

        Args:
            device_keys: The keys to be downloaded. A map from user ID, to a list of device IDs, or
                to an empty list to indicate all devices for the corresponding user.
            token: If the client is fetching keys as a result of a device update received in a sync
                request, this should be the 'since' token of that sync request, or any later sync
                token. This allows the server to ensure its response contains the keys advertised by
                the notification in that sync.
            timeout: The time (in milliseconds) to wait when downloading keys from remote servers.

        Returns:
            Information on the queried devices and errors for homeservers that could not be reached.
        """
        if isinstance(device_keys, (list, set)):
            device_keys = {user_id: [] for user_id in device_keys}
        data = {
            "timeout": timeout,
            "device_keys": device_keys,
        }
        if token:
            data["token"] = token
        resp = await self.api.request(Method.POST, Path.keys.query, data)
        return QueryKeysResponse.deserialize(resp)

    async def claim_keys(self, one_time_keys: Dict[UserID, Dict[DeviceID, EncryptionKeyAlgorithm]],
                         timeout: int = 10000) -> ClaimKeysResponse:
        """
        Claim one-time keys for use in pre-key messages.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.6.1#post-matrix-client-r0-keys-claim>`__

        Args:
            one_time_keys: The keys to be claimed. A map from user ID, to a map from device ID to
                algorithm name.
            timeout: The time (in milliseconds) to wait when downloading keys from remote servers.

        Returns:
            One-time keys for the queried devices and errors for homeservers that could not be
            reached.
        """
        resp = await self.api.request(Method.POST, Path.keys.claim, {
            "timeout": timeout,
            "one_time_keys": {user_id: {device_id: alg.serialize()
                                        for device_id, alg in devices.items()}
                              for user_id, devices in one_time_keys.items()},
        })
        return ClaimKeysResponse.deserialize(resp)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from attr import dataclass
import attr

from .primitive import ContentURI
from .util import SerializableAttrs


@dataclass
class MediaRepoConfig(SerializableAttrs['MediaRepoConfig']):
    """
    Matrix media repo config. See `GET /_matrix/media/r0/config`_.

    .. _GET /_matrix/media/r0/config:
        https://matrix.org/docs/spec/client_server/r0.5.0#get-matrix-media-r0-config
    """
    upload_size: int = attr.ib(metadata={"json": "m.upload.size"})


@dataclass
class OpenGraphImage(SerializableAttrs['OpenGraphImage']):
    url: ContentURI = attr.ib(default=None, metadata={"json": "og:image"})
    mimetype: str = attr.ib(default=None, metadata={"json": "og:image:type"})
    height: int = attr.ib(default=None, metadata={"json": "og:image:width"})
    width: int = attr.ib(default=None, metadata={"json": "og:image:height"})
    size: int = attr.ib(default=None, metadata={"json": "matrix:image:size"})


@dataclass
class OpenGraphVideo(SerializableAttrs['OpenGraphVideo']):
    url: ContentURI = attr.ib(default=None, metadata={"json": "og:video"})
    mimetype: str = attr.ib(default=None, metadata={"json": "og:video:type"})
    height: int = attr.ib(default=None, metadata={"json": "og:video:width"})
    width: int = attr.ib(default=None, metadata={"json": "og:video:height"})
    size: int = attr.ib(default=None, metadata={"json": "matrix:video:size"})


@dataclass
class OpenGraphAudio(SerializableAttrs['OpenGraphAudio']):
    url: ContentURI = attr.ib(default=None, metadata={"json": "og:audio"})
    mimetype: str = attr.ib(default=None, metadata={"json": "og:audio:type"})


@dataclass
class MXOpenGraph(SerializableAttrs['MXOpenGraph']):
    """
    Matrix URL preview response. See `GET /_matrix/media/r0/preview_url`_.

    .. _GET /_matrix/media/r0/preview_url:
        https://matrix.org/docs/spec/client_server/r0.5.0#get-matrix-media-r0-preview-url
    """
    title: str = attr.ib(default=None, metadata={"json": "og:title"})
    description: str = attr.ib(default=None, metadata={"json": "og:description"})
    image: OpenGraphImage = attr.ib(default=None, metadata={"flatten": True})
    video: OpenGraphVideo = attr.ib(default=None, metadata={"flatten": True})
    audio: OpenGraphAudio = attr.ib(default=None, metadata={"flatten": True})



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Set, Optional, Dict, TYPE_CHECKING
import logging
import asyncio

from mautrix.types import (RoomID, EventID, EventType, EventContent, EncryptedMegolmEventContent,
                           EncryptedEvent)
from mautrix.errors import MNotFound, EncryptionError, DecryptionError
from mautrix.util.logging import TraceLogger

from .store_updater import StoreUpdatingAPI
from .dispatcher import SimpleDispatcher

if TYPE_CHECKING:
    from mautrix.crypto import OlmMachine
    from .client import Client


class EncryptingAPI(StoreUpdatingAPI):
    _crypto: Optional['OlmMachine']
    encryption_blacklist: Set[EventType] = {EventType.REACTION}
    crypto_log: TraceLogger = logging.getLogger("mau.client.crypto")
    _share_session_events: Dict[RoomID, asyncio.Event]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._crypto = None
        self._share_session_events = {}

    @property
    def crypto(self) -> Optional['OlmMachine']:
        return self._crypto

    @crypto.setter
    def crypto(self, crypto: 'OlmMachine') -> None:
        if not self.state_store:
            raise ValueError("State store must be set to use encryption")
        self._crypto = crypto

    @property
    def crypto_enabled(self) -> bool:
        return bool(self.crypto) and bool(self.state_store)

    async def encrypt(self, room_id: RoomID, event_type: EventType, content: EventContent
                      ) -> EncryptedMegolmEventContent:
        try:
            return await self.crypto.encrypt_megolm_event(room_id, event_type, content)
        except EncryptionError:
            self.crypto_log.debug("Got EncryptionError, sharing group session and trying again")
            await self.share_group_session(room_id)
            self.crypto_log.trace(f"Shared group session, now trying to encrypt in {room_id} again")
            return await self.crypto.encrypt_megolm_event(room_id, event_type, content)

    async def _share_session_lock(self, room_id: RoomID) -> bool:
        try:
            event = self._share_session_events[room_id]
        except KeyError:
            self._share_session_events[room_id] = asyncio.Event()
            return True
        else:
            await event.wait()
            return False

    async def share_group_session(self, room_id: RoomID) -> None:
        if not await self._share_session_lock(room_id):
            self.log.silly("Group session was already being shared, so didn't share new one")
            return
        try:
            if not await self.state_store.has_full_member_list(room_id):
                self.crypto_log.trace(f"Don't have full member list for {room_id},"
                                      " fetching from server")
                members = list((await self.get_joined_members(room_id)).keys())
            else:
                self.crypto_log.trace(f"Fetching member list for {room_id} from state store")
                members = await self.state_store.get_members(room_id)
            await self.crypto.share_group_session(room_id, members)
        finally:
            self._share_session_events.pop(room_id).set()

    async def send_message_event(self, room_id: RoomID, event_type: EventType,
                                 content: EventContent, disable_encryption: bool = False,
                                 **kwargs) -> EventID:
        if self.crypto and event_type not in self.encryption_blacklist and not disable_encryption:
            is_encrypted = await self.state_store.is_encrypted(room_id)
            if is_encrypted is None:
                try:
                    await self.get_state_event(room_id, EventType.ROOM_ENCRYPTION)
                    is_encrypted = True
                except MNotFound:
                    is_encrypted = False
            if is_encrypted:
                content = await self.encrypt(room_id, event_type, content)
                event_type = EventType.ROOM_ENCRYPTED
        return await super().send_message_event(room_id, event_type, content, **kwargs)


class DecryptionDispatcher(SimpleDispatcher):
    event_type = EventType.ROOM_ENCRYPTED
    client: 'Client'

    async def handle(self, evt: EncryptedEvent) -> None:
        try:
            self.client.crypto_log.trace(f"Decrypting {evt.event_id} in {evt.room_id}...")
            decrypted = await self.client.crypto.decrypt_megolm_event(evt)
        except DecryptionError as e:
            self.client.crypto_log.warning(f"Failed to decrypt {evt.event_id}: {e}")
            return
        self.client.crypto_log.trace(f"Decrypted {evt.event_id}: {decrypted}")
        self.client.dispatch_event(decrypted, evt.source)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, TYPE_CHECKING

from mautrix.types import EventType, Event, StateEvent

from .state_store import SyncStore, StateStore
from .syncer import Syncer
from .encryption_manager import EncryptingAPI, DecryptionDispatcher

if TYPE_CHECKING:
    from mautrix.crypto import OlmMachine


class Client(EncryptingAPI, Syncer):
    """Client is a high-level wrapper around the client API."""

    def __init__(self, *args, sync_store: Optional[SyncStore] = None,
                 state_store: Optional[StateStore] = None, **kwargs) -> None:
        EncryptingAPI.__init__(self, *args, state_store=state_store, **kwargs)
        Syncer.__init__(self, sync_store)
        self.add_event_handler(EventType.ALL, self._update_state)

    async def _update_state(self, evt: Event) -> None:
        if not isinstance(evt, StateEvent) or not self.state_store:
            return
        await self.state_store.update_state(evt)

    @EncryptingAPI.crypto.setter
    def crypto(self, crypto: 'OlmMachine') -> None:
        if not self.state_store:
            raise ValueError("State store must be set to use encryption")
        self._crypto = crypto
        if self.crypto_enabled:
            self.add_dispatcher(DecryptionDispatcher)
        else:
            self.remove_dispatcher(DecryptionDispatcher)



# Copyright (c) 2021 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, Dict, Union, TYPE_CHECKING
from urllib.parse import quote as urllib_quote, urljoin as urllib_join
from json.decoder import JSONDecodeError
from enum import Enum
from time import time
import platform
import logging
import asyncio
import json

from yarl import URL
from aiohttp import ClientSession, __version__ as aiohttp_version
from aiohttp.client_exceptions import ContentTypeError, ClientError

from mautrix.errors import make_request_error, MatrixConnectionError, MatrixRequestError
from mautrix.util.logging import TraceLogger
from mautrix import __version__ as mautrix_version

if TYPE_CHECKING:
    from mautrix.types import JSON


class APIPath(Enum):
    """
    The known Matrix API path prefixes.
    These don't start with a slash so they can be used nicely with yarl.
    """
    CLIENT = "_matrix/client/r0"
    CLIENT_UNSTABLE = "_matrix/client/unstable"
    MEDIA = "_matrix/media/r0"

    def __repr__(self):
        return self.value

    def __str__(self):
        return self.value


class Method(Enum):
    """A HTTP method."""

    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"

    def __repr__(self):
        return self.value

    def __str__(self):
        return self.value


class PathBuilder:
    """
    A utility class to build API paths.

    Examples:
        >>> from mautrix.api import Path
        >>> room_id = "!foo:example.com"
        >>> event_id = "$bar:example.com"
        >>> str(Path.rooms[room_id].event[event_id])
        "_matrix/client/r0/rooms/%21foo%3Aexample.com/event/%24bar%3Aexample.com"
    """

    def __init__(self, path: Union[str, APIPath] = "") -> None:
        self.path: str = str(path)

    def __str__(self) -> str:
        return self.path

    def __repr__(self):
        return self.path

    def __getattr__(self, append: str) -> 'PathBuilder':
        if append is None:
            return self
        return PathBuilder(f"{self.path}/{append}")

    def raw(self, append: str) -> 'PathBuilder':
        """
        Directly append a string to the path.

        Args:
            append: The string to append.
        """
        if append is None:
            return self
        return PathBuilder(self.path + append)

    def __eq__(self, other: Union['PathBuilder', str]) -> bool:
        return other.path == self.path if isinstance(other, PathBuilder) else other == self.path

    @staticmethod
    def _quote(string: str) -> str:
        return urllib_quote(string, safe="")

    def __getitem__(self, append: Union[str, int]) -> 'PathBuilder':
        if append is None:
            return self
        return PathBuilder(f"{self.path}/{self._quote(str(append))}")


Path = PathBuilder(APIPath.CLIENT)
ClientPath = Path
UnstableClientPath = PathBuilder(APIPath.CLIENT_UNSTABLE)
MediaPath = PathBuilder(APIPath.MEDIA)

_req_id = 0


def next_global_req_id() -> int:
    global _req_id
    _req_id += 1
    return _req_id


class HTTPAPI:
    """HTTPAPI is a simple asyncio Matrix API request sender."""

    base_url: URL
    token: str
    log: TraceLogger
    loop: asyncio.AbstractEventLoop
    session: ClientSession
    txn_id: Optional[int]
    default_ua: str = (f"mautrix-python/{mautrix_version} aiohttp/{aiohttp_version} "
                       f"Python/{platform.python_version()}")
    global_default_retry_count: int = 0
    default_retry_count: int

    def __init__(self, base_url: Union[URL, str], token: str = "", *,
                 client_session: ClientSession = None, default_retry_count: int = None,
                 txn_id: int = 0, log: Optional[TraceLogger] = None,
                 loop: Optional[asyncio.AbstractEventLoop] = None) -> None:
        """
        Args:
            base_url: The base URL of the homeserver client-server API to use.
            token: The access token to use.
            client_session: The aiohttp ClientSession to use.
            txn_id: The outgoing transaction ID to start with.
            log: The logging.Logger instance to log requests with.
            default_retry_count: Default number of retries to do when encountering network errors.
        """
        self.base_url = URL(base_url)
        self.token = token
        self.log = log or logging.getLogger("mau.http")
        self.loop = loop or asyncio.get_event_loop()
        self.session = client_session or ClientSession(loop=self.loop,
                                                       headers={"User-Agent": self.default_ua})
        if txn_id is not None:
            self.txn_id = txn_id
        if default_retry_count is not None:
            self.default_retry_count = default_retry_count
        else:
            self.default_retry_count = self.global_default_retry_count

    async def _send(self, method: Method, url: URL, content: Union[bytes, str],
                    query_params: Dict[str, str], headers: Dict[str, str]) -> 'JSON':
        request = self.session.request(str(method), url, data=content,
                                       params=query_params, headers=headers)
        async with request as response:
            if response.status < 200 or response.status >= 300:
                errcode = message = None
                try:
                    response_data = await response.json()
                    errcode = response_data["errcode"]
                    message = response_data["error"]
                except (JSONDecodeError, ContentTypeError, KeyError):
                    pass
                raise make_request_error(http_status=response.status,
                                         text=await response.text(),
                                         errcode=errcode, message=message)
            return await response.json()

    def _log_request(self, method: Method, path: PathBuilder, content: Union[str, bytes],
                     orig_content, query_params: Dict[str, str], req_id: int) -> None:
        if not self.log:
            return
        log_content = content if not isinstance(content, bytes) else f"<{len(content)} bytes>"
        as_user = query_params.get("user_id", None)
        level = 1 if path == Path.sync else 5
        self.log.log(level, f"{method}#{req_id} /{path} {log_content}".strip(" "),
                     extra={"matrix_http_request": {
                         "req_id": req_id,
                         "method": str(method),
                         "path": str(path),
                         "content": (orig_content if isinstance(orig_content, (dict, list))
                                     else log_content),
                         "user": as_user,
                     }})

    def _full_path(self, path: Union[PathBuilder, str]) -> str:
        path = str(path)
        if path and path[0] == "/":
            path = path[1:]
        base_path = self.base_url.raw_path
        if base_path[-1] != "/":
            base_path += "/"
        return urllib_join(base_path, path)

    async def request(self, method: Method, path: Union[PathBuilder, str],
                      content: Optional[Union[dict, list, bytes, str]] = None,
                      headers: Optional[Dict[str, str]] = None,
                      query_params: Optional[Dict[str, str]] = None,
                      retry_count: Optional[int] = None) -> 'JSON':
        """
        Make a raw Matrix API request.

        Args:
            method: The HTTP method to use.
            path: The full API endpoint to call (including the _matrix/... prefix)
            content: The content to post as a dict/list (will be serialized as JSON)
                or bytes/str (will be sent as-is).
            headers: A dict of HTTP headers to send.
                If the headers don't contain ``Content-Type``, it'll be set to ``application/json``.
                The ``Authorization`` header is always overridden if :attr:`token` is set.
            query_params: A dict of query parameters to send.
            retry_count: Number of times to retry if the homeserver isn't reachable.

        Returns:
            The parsed response JSON.
        """
        headers = headers or {}
        if self.token:
            headers["Authorization"] = f"Bearer {self.token}"
        query_params = query_params or {}

        if method != Method.GET:
            content = content or {}
            if "Content-Type" not in headers:
                headers["Content-Type"] = "application/json"
            orig_content = content
            is_json = headers.get("Content-Type", None) == "application/json"
            if is_json and isinstance(content, (dict, list)):
                content = json.dumps(content)
        else:
            orig_content = content = None
        full_url = self.base_url.with_path(self._full_path(path), encoded=True)
        req_id = next_global_req_id()

        if retry_count is None:
            retry_count = self.default_retry_count
        backoff = 4
        while True:
            self._log_request(method, path, content, orig_content, query_params, req_id)
            try:
                return await self._send(method, full_url, content, query_params, headers or {})
            except MatrixRequestError as e:
                if retry_count > 0 and e.http_status in (502, 503, 504):
                    self.log.warning(f"Request #{req_id} failed with HTTP {e.http_status}, "
                                     f"retrying in {backoff} seconds")
                else:
                    raise
            except ClientError as e:
                if retry_count > 0:
                    self.log.warning(f"Request #{req_id} failed with {e}, "
                                     f"retrying in {backoff} seconds")
                else:
                    raise MatrixConnectionError(str(e)) from e
            await asyncio.sleep(backoff)
            backoff *= 2
            retry_count -= 1

    def get_txn_id(self) -> str:
        """Get a new unique transaction ID."""
        self.txn_id += 1
        return f"mautrix-python_R{self.txn_id}@T{int(time() * 1000)}"

    def get_download_url(self, mxc_uri: str, download_type: str = "download") -> URL:
        """
        Get the full HTTP URL to download a mxc:// URI.

        Args:
            mxc_uri: The MXC URI whose full URL to get.
            download_type: The type of download ("download" or "thumbnail")

        Returns:
            The full HTTP URL.

        Raises:
            ValueError: If `mxc_uri` doesn't begin with mxc://

        Examples:
            >>> api = HTTPAPI(...)
            >>> api.get_download_url("mxc://matrix.org/pqjkOuKZ1ZKRULWXgz2IVZV6")
            "https://matrix.org/_matrix/media/r0/download/matrix.org/pqjkOuKZ1ZKRULWXgz2IVZV6"
        """
        if mxc_uri.startswith("mxc://"):
            return self.base_url / str(APIPath.MEDIA) / download_type / mxc_uri[6:]
        else:
            raise ValueError("MXC URI did not begin with `mxc://`")
